/***************************************************************************
                           testdialog.cpp
                           -------------------
    copyright            : (C) 2004 by Jos van den Oever
    email                : jos@vandenoever.info
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include "testdialog.h"
#include "cubedialog.h"
#include "scorekeeper.h"
#include "sidelib.h"
#include "filesidestyle.h"
#include <QFileInfo>
#include <QApplication>
#include <QDir>
#include <QLabel>
#include <QPushButton>
#include <QComboBox>
#include <QTextBrowser>
#include <QPixmap>
#include <QKeyEvent>
#include <QResizeEvent>
#include <QMouseEvent>
#include <QVBoxLayout>
#include <iostream>
using std::cerr;
using std::endl;

class ClickableQLabel : public QLabel {
public:
	ClickableQLabel(QWidget* parent=0, Qt::WFlags f=0);
	void setAlternative(const QString &);
protected:
	void mouseReleaseEvent(QMouseEvent *e);
private:
	QString alternative;
};
ClickableQLabel::ClickableQLabel(QWidget* parent, Qt::WFlags f)
		: QLabel(parent,f) {setWordWrap(true);}
void
ClickableQLabel::setAlternative(const QString &s) {
	alternative = s;
}
void
ClickableQLabel::mouseReleaseEvent(QMouseEvent *) {
	QString temp = text();
	setText(alternative);
	alternative = temp;
}

TestDialog::TestDialog(QWidget* parent, Qt::WFlags f)
		: QWidget(parent, f) {
	setLayout(new QVBoxLayout());
	layout()->setMargin(0);

	setBackgroundPixmap("background.jpg");

	welcome = new ClickableQLabel();
	layout()->addWidget(welcome);
	welcome->setText(QString("<p><b>")+tr("Welcome to the cube test.") +"</b></p><p>"+tr("We are going to compare cubes.") +"</p><p>"+tr("Choose the cube that is the same as the" " top cube. You can turn the cubes with " " the mouse. The six squares next to the "
			 "cube are the same as the six sides of the "
			 "cube.")
			 +"</p><p>"+tr("Click \"Start\" to begin.")+"</p>");
	welcome->setAlternative(tr("CubeTest is written by Jos van den Oever and "
				   "released under the General Public License."));

	welcome->setAlignment(Qt::AlignHCenter | Qt::AlignVCenter);
			 //     | Qt::WordBreak);
	conclusion = new QLabel("");
	conclusion->setWordWrap(true);
	layout()->addWidget(conclusion);
	conclusion->setAlignment(Qt::AlignHCenter | Qt::AlignVCenter);
	conclusion->hide();

	options = new QWidget();
	layout()->addWidget(options);
	QVBoxLayout *optionslayout = new QVBoxLayout();
	options->setLayout(optionslayout);
	options->setSizePolicy(QSizePolicy(QSizePolicy::Preferred,
					   QSizePolicy::Fixed));
	QHBoxLayout *row = new QHBoxLayout();
	optionslayout->addLayout(row);
	QLabel *l = new QLabel(tr("Difficulty "));
	row->addWidget(l);
	l->setAlignment(Qt::AlignRight | Qt::AlignVCenter);
	QComboBox *levels = new QComboBox();
	row->addWidget(levels);
	levels->addItem(tr("Easy"));
	levels->addItem(tr("Medium"));
	levels->addItem(tr("Difficult"));
	levels->addItem(tr("Challenging"));
	connect(levels, SIGNAL(activated(int)), this, SLOT(setLevel(int)));
	fss = new FileSideStyle();
	if (fss->getStyles().size() > 1) {
		row = new QHBoxLayout();
		optionslayout->addLayout(row);
		l = new QLabel(tr("Style "));
		row->addWidget(l);
		l->setAlignment(Qt::AlignRight | Qt::AlignVCenter);
		QComboBox *stylelist = new QComboBox();
		row->addWidget(stylelist);
		stylelist->addItems(fss->getStyles());
		connect(stylelist, SIGNAL(activated(const QString &)),
			this, SLOT(changeStyle(const QString &)));
		fss->loadStyle(stylelist->currentText());
		ss = fss;
	} else if (fss->getStyles().size() == 1) {
		fss->loadStyle(fss->getStyles()[0]);
		ss = fss;
	} else {
		delete fss;
		fss = 0;
		ss = new SideLib();
	}
	start = new QPushButton(tr("Start"));
	layout()->addWidget(start);
	connect(start, SIGNAL(clicked()), this, SLOT(startQuestions()));
	bugmail = new QLabel(tr("Please send suggestions, themes, translations,"
				" and bugsreports to jos@vandenoever.info"));
	layout()->addWidget(bugmail);
	bugmail->setSizePolicy(QSizePolicy(QSizePolicy::Ignored,
					   QSizePolicy::Fixed));

	cd = new CubeDialog();
	layout()->addWidget(cd);
	connect(cd, SIGNAL(finished()), this, SLOT(finishQuestions()));
	cd->setSideStyle(ss);
	cd->hide();
	setLevel(0);
	start->setFocus();
}
TestDialog::~TestDialog() {
	delete ss;
}
void
TestDialog::startQuestions() {
	welcome->hide();
	options->hide();
	start->hide();
	conclusion->hide();
	start->hide();
	bugmail->hide();
	cd->startQuestions();
	cd->show();
}
void
TestDialog::setQuestion(Question &q) {
	cd->setQuestion(q);
	startQuestions();
}
void
TestDialog::finishQuestions() {
	cd->hide();
	QString message
	= QString(tr("You've answered %1 out of %2 questions correctly."))
	  .arg(QString::number(cd->getScore().getPoints()))
	  .arg(QString::number(cd->getScore().getQuestions()));
	conclusion->setText(message);
	conclusion->show();
	start->show();
	options->show();
	bugmail->show();
	start->setFocus();
}
void
TestDialog::setBackgroundPixmap(const QString &s) {
	QPixmap qpm = QPixmap(qApp->applicationDirPath()
		+ "/../share/cubetest/images/"+s);
	if (qpm.isNull()) return;
	QPalette pal = palette();
	QBrush b = pal.background();
	b.setTexture(qpm);
	pal.setBrush(QPalette::Background, b);
	setPalette(pal);
}
void
TestDialog::changeStyle(const QString &s) {
	if (fss) {
		fss->loadStyle(s);
	}
}
void
TestDialog::keyReleaseEvent(QKeyEvent *e) {
	if (e->key() == Qt::Key_Escape) {
		finishQuestions();
	} else if (e->key() == Qt::Key_F1) {
		QTextBrowser *doc = new QTextBrowser();
		doc->setSource(qApp->applicationDirPath()
			+ "/../share/cubetest/doc/readme.html");
		doc->resize(600,400);
		doc->setWindowTitle("http://www.vandenoever.info/cubetest/");
		doc->show();
	} else {
		e->ignore();
	}
}
void
TestDialog::resizeEvent(QResizeEvent *e) {
	// adjust the fonts in rough steps to fit to the widget size
	int size = (width()>height()) ? height() : width();
	int fontsize = size/32;
	if (fontsize < 12)
		fontsize = 12;
	QFont f = font();
	f.setPixelSize(fontsize);
	setFont(f);
	QWidget::resizeEvent(e);
}
void
TestDialog::setLevel(int i) {
	cd->blanks = 3-i;
}
